# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
from typing import Tuple
from dataclasses import dataclass, field

# Default SDF version used in the serialized XML context
SDF_VERSION=1.7

# Read the following for more information about dataclasses internals:
# https://florimond.dev/blog/articles/2018/10/reconciling-dataclasses-and-properties-in-python/


@dataclass
class GazeboPlugin(abc.ABC):
    """
    Base class of all World and Model plugins for Ignition Gazebo.

    The Plugin abstract class provides boilerplate code that simplifies and streamlines
    the definition of helper classes that insert Ignition Gazebo plugins to either World
    or Model objects.

    Classes that inherit from Plugin have to provide the following information:

    1) All the properties of the plugin as dataclass fields
    2) The private specification of the plugin (plugin name and plugin class)
    3) Optionally: the serialized XML context

    Example:

    .. code-block:: python

        plugin = MyPlugin(my_property=42)
        model = MyModel(world=my_world)

        model.insert_model_plugin(*plugin.args())
    """

    _plugin_name: str = field(init=False, repr=False)
    _plugin_class: str = field(init=False, repr=False)

    def to_xml(self) -> str:
        """
        Get the XML plugin content.

        Returns:
            The XML plugin content.
        """
        return ""

    def args(self) -> Tuple[str, str, str]:
        """
        Get the arguments passed to the ScenarI/O methods used to insert plugins.

        Returns:
            A tuple with the args required to insert the plugin.
        """
        return str(self._plugin_name), \
               str(self._plugin_class), \
               GazeboPlugin.wrap_in_sdf(self.to_xml())

    @staticmethod
    def wrap_in_sdf(context: str) -> str:
        """
        Wrap the XML context inside a SDF root.

        Args:
            context: The XML string with the plugin's context.

        Returns:
            The plugin's context wrapped inside a SDF root.
        """

        return f"""<sdf version='{SDF_VERSION}'>{context}</sdf>"""



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import gym.spaces
import numpy as np
from typing import List, Tuple, Dict, Union, NewType

Done = NewType('Done', bool)
Info = NewType('Info', Dict)
Reward = NewType('Reward', float)
Observation = NewType('Observation', np.ndarray)
Action = NewType('Action', Union[np.ndarray, np.number])

SeedList = NewType('SeedList', List[int])

State = NewType('State', Tuple[Observation, Reward, Done, Info])

ActionSpace = NewType('ActionSpace', gym.spaces.Space)
ObservationSpace = NewType('ObservationSpace', gym.spaces.Space)



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import gym
import pytest
import numpy as np
from gym.envs import registry
from gym_ignition.utils import logger
from gym.envs.registration import register
from gym_ignition.robots.sim import gazebo, pybullet
from gym_ignition.tasks.pendulum_swingup import PendulumSwingUp
from gym_ignition.utils.typing import Observation, Reward, State

# Set verbosity
logger.set_level(gym.logger.DEBUG)


class PendulumEnv(gym.Env):
    """
    Environment that implements the pendulum dynamics from equations.
    It integrates the system with Euler.
    """
    metadata = {'render.modes': []}

    def __init__(self):
        super().__init__()

        # Check the xacro pendulum model
        r = 0.01
        self.L = 0.5
        self.m = 1
        self.g = 9.8182
        self.I = self.m * (4 * self.L * self.L + 3 * r * r) / 12

        self.dt = None
        # self.force = None
        self.theta = None
        self.theta_dot = None

    def set_state_from_obs(self, observation: np.ndarray) -> None:
        """
        Set the state of the environment from an observation sampled from another
        environment.
        """

        # Unpack the observation
        cos_theta, sin_theta, self.theta_dot = observation.tolist()

        # Calculate the initial angle
        self.theta = np.math.atan2(sin_theta, cos_theta)

    def step(self, action: np.ndarray) -> State:
        tau_m = action[0]
        theta_ddot = \
            (self.m * self.g * self.L / 2.0 * np.sin(self.theta) + tau_m) / self.I

        # Integrate with euler.
        # Note that in this way the ground truth used as comparison implements a very
        # basic integration method, and the errors might be reduced implementing a
        # better integrator.
        self.theta_dot = self.theta_dot + theta_ddot * self.dt
        self.theta = self.theta + self.theta_dot * self.dt

        observation = np.array([np.cos(self.theta), np.sin(self.theta), self.theta_dot])

        return State((Observation(observation), Reward(0.0), False, {}))

    def reset(self):
        # Use set_state_from_obs
        pass

    def render(self, mode='human', **kwargs):
        raise Exception("This runtime does not support rendering")

    def seed(self, seed=None):
        raise Exception("This runtime should not be seeded")


def theta_from_obs(observation: np.ndarray) -> float:
    cos_theta, sin_theta, _ = observation
    return np.math.atan2(sin_theta, cos_theta)


if "Pendulum-Ignition-PyTest-v0" not in [spec.id for spec in list(registry.all())]:
    register(
        id='Pendulum-Ignition-PyTest-v0',
        entry_point='gym_ignition.runtimes.gazebo_runtime:GazeboRuntime',
        max_episode_steps=1000,
        kwargs={'task_cls': PendulumSwingUp,
                'robot_cls': gazebo.pendulum.PendulumGazeboRobot,
                'model': "Pendulum/Pendulum.urdf",
                'world': "DefaultEmptyWorld.world",
                'rtf': 100,
                'agent_rate': 4000,
                'physics_rate': 4000,
                'hard_reset': False,
                })

if "Pendulum-PyBullet-PyTest-v0" not in [spec.id for spec in list(registry.all())]:
    register(
        id='Pendulum-PyBullet-PyTest-v0',
        entry_point='gym_ignition.runtimes.pybullet_runtime:PyBulletRuntime',
        max_episode_steps=1000,
        kwargs={'task_cls': PendulumSwingUp,
                'robot_cls': pybullet.pendulum.PendulumPyBulletRobot,
                'model': "Pendulum/Pendulum.urdf",
                'world': "plane_implicit.urdf",
                'rtf': 100,
                'agent_rate': 4000,
                'physics_rate': 4000,  # To keep errors small, pybullet needs higher rate
                'hard_reset': False,
                })


def template_pendulum_wrt_ground_truth(env_name: str, max_error_in_deg: float):
    # Create the pendulum
    env = gym.make(env_name)

    # Create the environment with the equations and initialize its time step
    env_equation = PendulumEnv()
    env_equation.dt = 1.0 / env.unwrapped.spec._kwargs['agent_rate']

    # Render the environment
    # env.render('human')
    # time.sleep(5)

    # Seed the environment
    env.seed(42)

    for epoch in range(10):
        # Reset the environment
        logger.info("Resetting the environment")
        observation = env.reset()
        env_equation.set_state_from_obs(observation)

        # Initialize intermediate variables
        iteration = 0
        done = False

        while not done:
            iteration += 1

            # Sample a random action from the environment
            action = env.action_space.sample()

            # Step the environments
            observation, _, done, _ = env.step(action)
            observation_equation, _, _, _ = env_equation.step(action)

            theta = np.rad2deg(theta_from_obs(observation))
            theta_equation = np.rad2deg(theta_from_obs(observation_equation))

            # Compute the error taking care of the change of sign
            if np.sign(theta_equation) * np.sign(theta) == -1:
                error = (180 % abs(theta)) + (180 % abs(theta_equation))
            else:
                error = abs(theta - theta_equation)

            print(iteration, error)

            if error > max_error_in_deg:
                print("===================")
                print(f"Environment name: {env_name}")
                print(f"Iteration: #{iteration}")
                print(f"Error: {error}")
                print(f"Theta Equation    (deg): {theta_equation}")
                print(f"Theta Environment (deg): {theta}")
                print("===================")
                assert False, "Error in pendulum angle is bigger then the threshold"

    env.close()


@pytest.mark.parametrize("env_name, max_error_in_deg",
                         [("Pendulum-Ignition-PyTest-v0", 3.0),
                          ("Pendulum-PyBullet-PyTest-v0", 3.0),
                          ])
def test_pendulum_ignition(env_name: str, max_error_in_deg: float):
    template_pendulum_wrt_ground_truth(env_name, max_error_in_deg)



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

# Workaround for https://github.com/osrf/sdformat/issues/227.
# It has to be done before loading the bindings.
import gym_ignition_models
gym_ignition_models.setup_environment()

# Add IGN_GAZEBO_RESOURCE_PATH to the default search path
import os
from gym_ignition.utils import resource_finder
if "IGN_GAZEBO_RESOURCE_PATH" in os.environ:
    resource_finder.add_path_from_env_var("IGN_GAZEBO_RESOURCE_PATH")



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
from scenario import core as scenario


class ModelWrapper(scenario.Model, abc.ABC):

    def __init__(self, model: scenario.Model):

        # No need to call scenario.Model.__init__()!
        abc.ABC.__init__(self)

        self.model = model

    def __getattr__(self, name):

        return getattr(self.model, name)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest
pytestmark = pytest.mark.gym_ignition

from lxml import etree
import gym_ignition_models
from gym_ignition.utils import misc

from scenario import gazebo as scenario
from gym_ignition.randomizers.model import sdf
from gym_ignition.randomizers.model.sdf import Distribution, Method
from gym_ignition.randomizers.model.sdf import UniformParams, GaussianParams


def test_sdf_randomizer():

    # Get the URDF model
    urdf_model = gym_ignition_models.get_model_file("cartpole")

    # Convert it to a SDF string
    sdf_model_string = scenario.urdffile_to_sdfstring(urdf_model)

    # Write the SDF string to a temp file
    sdf_model = misc.string_to_file(sdf_model_string)

    # Create the randomizer
    randomizer = sdf.SDFRandomizer(sdf_model=sdf_model)

    # Get the original model string. It is parsed and then serialized without changes.
    orig_model = randomizer.sample(pretty_print=True)

    with pytest.raises(ValueError):
        # Setting wrong distribution
        randomizer.new_randomization() \
            .at_xpath("*/link[@name='pole']/inertial/inertia/ixx") \
            .method(Method.Additive) \
            .sampled_from(Distribution.Uniform, GaussianParams(mean=0, variance=0.1)) \
            .add()

    # Add a uniform randomization
    randomizer.new_randomization() \
        .at_xpath("*/link[@name='pole']/inertial/inertia/ixx") \
        .method(Method.Additive) \
        .sampled_from(Distribution.Gaussian, GaussianParams(mean=0, variance=0.1)) \
        .add()

    randomizer.process_data()
    assert len(randomizer.get_active_randomizations()) == 1

    assert randomizer.sample(pretty_print=True) != orig_model

    # Clean the randomizer
    randomizer.clean()
    assert len(randomizer.get_active_randomizations()) == 0
    assert randomizer.sample(pretty_print=True) == orig_model

    # Add a multi-match randomization
    randomizer.new_randomization() \
        .at_xpath("*/link/inertial/inertia/ixx") \
        .method(Method.Coefficient) \
        .sampled_from(Distribution.Uniform, UniformParams(low=0.8, high=1.2)) \
        .add()

    assert len(randomizer.get_active_randomizations()) == 1

    # Expand the matches
    randomizer.process_data()
    assert len(randomizer.get_active_randomizations()) > 1

    # Sample
    assert randomizer.sample(pretty_print=True) != orig_model


def test_randomizer_reproducibility():

    # Get the model
    sdf_model = gym_ignition_models.get_model_file("ground_plane")

    # Initialize the randomizers
    randomizer1 = sdf.SDFRandomizer(sdf_model=sdf_model)
    randomizer2 = sdf.SDFRandomizer(sdf_model=sdf_model)
    randomizer3 = sdf.SDFRandomizer(sdf_model=sdf_model)

    # Randomize the ground friction of all links (the ground plane collision)
    frictions = randomizer1.find_xpath("*/link/collision/surface/friction")
    assert len(frictions) == 1

    # Get the original model string. It is parsed and then serialized without changes.
    orig_model_string1 = randomizer1.sample(pretty_print=True)
    orig_model_string2 = randomizer2.sample(pretty_print=True)
    orig_model_string3 = randomizer3.sample(pretty_print=True)
    assert orig_model_string1 == orig_model_string2 == orig_model_string3

    # Do not seed #3
    randomizer1.seed(42)
    randomizer2.seed(42)

    # Add randomizations for #1
    randomizer1.new_randomization() \
        .at_xpath("*/link/collision/surface/friction/ode/mu") \
        .method(Method.Absolute) \
        .sampled_from(Distribution.Uniform,
                      UniformParams(low=0, high=100)) \
        .add()

    # Add randomizations for #2
    randomizer2.new_randomization() \
        .at_xpath("*/link/collision/surface/friction/ode/mu") \
        .method(Method.Absolute) \
        .sampled_from(Distribution.Uniform,
                      UniformParams(low=0, high=100)) \
        .add()

    # Add randomizations for #3
    randomizer3.new_randomization() \
        .at_xpath("*/link/collision/surface/friction/ode/mu") \
        .method(Method.Absolute) \
        .sampled_from(Distribution.Uniform,
                      UniformParams(low=0, high=100)) \
        .add()

    # Process the randomizations
    randomizer1.process_data()
    randomizer2.process_data()
    randomizer3.process_data()

    for _ in range(5):
        model1 = randomizer1.sample()
        model2 = randomizer2.sample()
        model3 = randomizer3.sample()

        assert model1 == model2
        assert model1 != model3


def test_randomize_missing_element():

    # Get the URDF model
    urdf_model = gym_ignition_models.get_model_file("pendulum")

    # Convert it to a SDF string
    sdf_model_string = scenario.urdffile_to_sdfstring(urdf_model)

    # Write the SDF string to a temp file
    sdf_model = misc.string_to_file(sdf_model_string)

    # Create the randomizer
    randomizer = sdf.SDFRandomizer(sdf_model=sdf_model)

    # Try to randomize a missing element
    with pytest.raises(RuntimeError):
        # The ode/mu elements are missing
        randomizer.new_randomization() \
            .at_xpath("*/link/collision/surface/friction/ode/mu") \
            .method(Method.Absolute) \
            .sampled_from(Distribution.Uniform,
                          UniformParams(low=0, high=100)) \
            .add()

    # Add the missing friction/ode/mu element. We assume that friction exists.
    frictions = randomizer.find_xpath("*/link/collision/surface/friction")

    for friction in frictions:

        # Create parent 'ode' first
        if friction.find("ode") is None:
            etree.SubElement(friction, "ode")

        # Create child 'mu' after
        ode = friction.find("ode")
        if ode.find("mu") is None:
            etree.SubElement(ode, "mu")

        # Assign a dummy value to mu
        mu = ode.find("mu")
        mu.text = str(0)

    # Apply the same randomization
    randomizer.new_randomization() \
        .at_xpath("*/link/collision/surface/friction/ode/mu") \
        .method(Method.Absolute) \
        .sampled_from(Distribution.Uniform,
                      UniformParams(low=0, high=100)) \
        .ignore_zeros(False) \
        .add()

    # Process the randomization and sample a model
    randomizer.process_data()

    model1 = randomizer.sample(pretty_print=True)
    model2 = randomizer.sample(pretty_print=True)
    assert model1 != model2


def test_full_panda_randomization():

    # Get the URDF model
    urdf_model = gym_ignition_models.get_model_file("panda")

    # Convert it to a SDF string
    sdf_model_string = scenario.urdffile_to_sdfstring(urdf_model)

    # Write the SDF string to a temp file
    sdf_model = misc.string_to_file(sdf_model_string)

    # Create the randomizer
    randomizer = sdf.SDFRandomizer(sdf_model=sdf_model)

    joint_dynamics = randomizer.find_xpath("*/joint/axis/dynamics")
    assert len(joint_dynamics) > 0

    # Add the friction and damping elements since they're missing in the model
    for joint_dynamic in joint_dynamics:

        if joint_dynamic.find("friction") is None:
            etree.SubElement(joint_dynamic, "friction")
            friction = joint_dynamic.find("friction")
            friction.text = str(0)

        if joint_dynamic.find("damping") is None:
            etree.SubElement(joint_dynamic, "damping")
            damping = joint_dynamic.find("damping")
            damping.text = str(3)

    randomization_config = {
        "*/link/inertial/mass": {
            # mass + U(-0.5, 0.5)
            'method': Method.Additive,
            'distribution': Distribution.Uniform,
            'params': UniformParams(low=-0.5, high=0.5),
            'ignore_zeros': True,
            'force_positive': True,
        },
        "*/link/inertial/inertia/ixx": {
            # inertia * N(1, 0.2)
            'method': Method.Coefficient,
            'distribution': Distribution.Gaussian,
            'params': GaussianParams(mean=1.0, variance=0.2),
            'ignore_zeros': True,
            'force_positive': True,
        },
        "*/link/inertial/inertia/iyy": {
            'method': Method.Coefficient,
            'distribution': Distribution.Gaussian,
            'params': GaussianParams(mean=1.0, variance=0.2),
            'ignore_zeros': True,
            'force_positive': True,
        },
        "*/link/inertial/inertia/izz": {
            'method': Method.Coefficient,
            'distribution': Distribution.Gaussian,
            'params': GaussianParams(mean=1.0, variance=0.2),
            'ignore_zeros': True,
            'force_positive': True,
        },
        "*/joint/axis/dynamics/friction": {
            # friction in [0, 5]
            'method': Method.Absolute,
            'distribution': Distribution.Uniform,
            'params': UniformParams(low=0, high=5),
            'ignore_zeros': False,  # We initialized the value as 0
            'force_positive': True,
        },
        "*/joint/axis/dynamics/damping": {
            # damping (= 3.0) * [0.8, 1.2]
            'method': Method.Coefficient,
            'distribution': Distribution.Uniform,
            'params': UniformParams(low=0.8, high=1.2),
            'ignore_zeros': True,
            'force_positive': True,
        },
        # TODO: */joint/axis/limit/effort
    }

    for xpath, config in randomization_config.items():

        randomizer.new_randomization() \
            .at_xpath(xpath) \
            .method(config["method"]) \
            .sampled_from(config["distribution"], config['params']) \
            .force_positive(config["distribution"]) \
            .ignore_zeros(config["ignore_zeros"]) \
            .add()

    randomizer.process_data()
    assert len(randomizer.get_active_randomizations()) > 0

    randomizer.sample(pretty_print=True)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest
pytestmark = pytest.mark.scenario

import numpy as np
from scenario import core
import gym_ignition_models
from ..common import utils
from scenario import gazebo as scenario
from gym_ignition.context.gazebo import controllers
from ..common.utils import gazebo_fixture as gazebo

# Set the verbosity
scenario.set_verbosity(scenario.Verbosity_debug)


@pytest.mark.parametrize("gazebo",
                         [(0.001, 5.0, 1)],
                         indirect=True,
                         ids=utils.id_gazebo_fn)
def test_computed_torque_fixed_base(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()
    step_size = gazebo.step_size()

    # Get the default world
    world = gazebo.get_world()

    # Insert the physics
    assert world.set_physics_engine(scenario.PhysicsEngine_dart)

    # Get the panda urdf
    panda_urdf = gym_ignition_models.get_model_file("panda")

    # Insert the panda arm
    model_name = "panda"
    assert world.insert_model(panda_urdf, core.Pose_identity(), model_name)

    # import time
    # gazebo.gui()
    # time.sleep(3)
    # gazebo.run(paused=True)

    # Get the model
    panda = world.get_model(model_name).to_gazebo()

    # Set the controller period
    panda.set_controller_period(step_size)

    # Insert the controller
    assert panda.insert_model_plugin(*controllers.ComputedTorqueFixedBase(
        kp=[10.0] * panda.dofs(),
        ki=[0.0] * panda.dofs(),
        kd=[3.0] * panda.dofs(),
        urdf=panda_urdf,
        joints=panda.joint_names(),
    ).args())

    # Set the references
    assert panda.set_joint_position_targets([0.0] * panda.dofs())
    assert panda.set_joint_velocity_targets([0.0] * panda.dofs())
    assert panda.set_joint_acceleration_targets([0.0] * panda.dofs())

    joints_no_fingers = [j for j in panda.joint_names() if j.startswith("panda_joint")]
    nr_of_joints = len(joints_no_fingers)
    assert nr_of_joints > 0

    # Reset the joints state
    q0 = [np.deg2rad(45)] * nr_of_joints
    dq0 = [0.1] * nr_of_joints
    assert panda.reset_joint_positions(q0, joints_no_fingers)
    assert panda.reset_joint_velocities(dq0, joints_no_fingers)

    assert gazebo.run(True)
    assert panda.joint_positions(joints_no_fingers) == pytest.approx(q0)
    assert panda.joint_velocities(joints_no_fingers) == pytest.approx(dq0)

    # Step the simulator for a couple of seconds
    for _ in range(3000):
        gazebo.run()

    # Check that the the references have been reached
    assert panda.joint_positions() == pytest.approx(panda.joint_position_targets(),
                                                    abs=np.deg2rad(1))
    assert panda.joint_velocities() == pytest.approx(panda.joint_velocity_targets(),
                                                     abs=0.05)

    # Apply an external force
    assert panda.get_link("panda_link4").to_gazebo().apply_world_force([100.0, 0, 0], 0.5)

    for _ in range(4000):
        assert gazebo.run()

    # Check that the the references have been reached
    assert panda.joint_positions() == pytest.approx(panda.joint_position_targets(),
                                                    abs=np.deg2rad(1))
    assert panda.joint_velocities() == pytest.approx(panda.joint_velocity_targets(),
                                                     abs=0.05)



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import gym
import pytest
import numpy as np
from gym.envs import registry
from gym_ignition.utils import logger
from gym.envs.registration import register
from gym_ignition.robots.sim import gazebo, pybullet
from gym_ignition.tasks.pendulum_swingup import PendulumSwingUp
from gym_ignition.tasks.cartpole_discrete import CartPoleDiscrete

# Set verbosity
logger.set_level(gym.logger.DEBUG)

if "Pendulum-Ignition-PyTest-v0" not in [spec.id for spec in list(registry.all())]:
    register(
        id='Pendulum-Ignition-PyTest-v0',
        entry_point='gym_ignition.runtimes.gazebo_runtime:GazeboRuntime',
        max_episode_steps=1000,
        kwargs={'task_cls': PendulumSwingUp,
                'robot_cls': gazebo.pendulum.PendulumGazeboRobot,
                'model': "Pendulum/Pendulum.urdf",
                'world': "DefaultEmptyWorld.world",
                'rtf': 100,
                'agent_rate': 4000,
                'physics_rate': 4000,
                'hard_reset': False,
                })

if "Pendulum-PyBullet-PyTest-v0" not in [spec.id for spec in list(registry.all())]:
    register(
        id='Pendulum-PyBullet-PyTest-v0',
        entry_point='gym_ignition.runtimes.pybullet_runtime:PyBulletRuntime',
        max_episode_steps=1000,
        kwargs={'task_cls': PendulumSwingUp,
                'robot_cls': pybullet.pendulum.PendulumPyBulletRobot,
                'model': "Pendulum/Pendulum.urdf",
                'world': "plane_implicit.urdf",
                'rtf': 100,
                'agent_rate': 4000,
                'physics_rate': 4000,  # To keep errors small, pybullet needs a higher rate
                'hard_reset': False,
                })

if "CartPoleDiscrete-Ignition-PyTest-v0" not in [spec.id for spec in list(registry.all())]:
    register(
        id='CartPoleDiscrete-Ignition-PyTest-v0',
        entry_point='gym_ignition.runtimes.gazebo_runtime:GazeboRuntime',
        max_episode_steps=500,
        kwargs={'task_cls': CartPoleDiscrete,
                'robot_cls': gazebo.cartpole.CartPoleGazeboRobot,
                'model': "CartPole/CartPole.urdf",
                'world': "DefaultEmptyWorld.world",
                'rtf': 100,
                'agent_rate': 4000,
                'physics_rate': 4000,
                'hard_reset': False,
                })

if "CartPoleDiscrete-PyBullet-PyTest-v0" not in [spec.id for spec in list(registry.all())]:
    register(
        id='CartPoleDiscrete-PyBullet-PyTest-v0',
        entry_point='gym_ignition.runtimes.pybullet_runtime:PyBulletRuntime',
        max_episode_steps=500,
        kwargs={'task_cls': CartPoleDiscrete,
                'robot_cls': pybullet.cartpole.CartPolePyBulletRobot,
                'model': "CartPole/CartPole.urdf",
                'world': "plane_implicit.urdf",
                'rtf': 100,
                'agent_rate': 4000,
                'physics_rate': 4000,
                'hard_reset': False,
                })


def template_compare_environments(env_name_a: str, env_name_b: str, max_error: float):
    # Create the pendulum
    env_a = gym.make(env_name_a)
    env_b = gym.make(env_name_b)

    assert env_a.unwrapped.spec._kwargs['agent_rate'] == \
           env_b.unwrapped.spec._kwargs['agent_rate']

    logger.set_level(gym.logger.DEBUG)

    # Render the environment
    # env.render('human')
    # time.sleep(5)

    # Seed the environment
    env_a.seed(42)
    env_b.seed(42)

    range_obs = env_a.observation_space.high - env_a.observation_space.low
    range_obs = np.where(np.isinf(range_obs), 1000, range_obs)

    for epoch in range(10):
        # Reset the environments
        observation_a = env_a.reset()
        observation_b = env_b.reset()

        assert np.allclose(observation_a, observation_b), \
            "Observations after reset don't match"

        # Initialize intermediate variables
        iteration = 0
        done_a = False

        while not done_a:
            iteration += 1

            # Sample a random action from the environment a
            action = env_a.action_space.sample()

            # Step the environments
            observation_a, _, done_a, _ = env_b.step(action)
            observation_b, _, done_b, _ = env_a.step(action)

            error = np.sum(np.abs(observation_a / range_obs - observation_b / range_obs))

            if error > max_error:
                print("===================")
                print(f"Environment A name: {env_name_a}")
                print(f"Environment B name: {env_name_b}")
                print(f"Rollout: #{epoch}@{iteration}")
                print(f"Error: {error}")
                print(f"Max Error: {max_error}")
                print(f"Observation A: {observation_a}")
                print(f"Observation B: {observation_b}")
                print("===================")
                assert False, "The error is bigger than the threshold"

    env_a.close()
    env_b.close()


@pytest.mark.parametrize("env_name_a, env_name_b, max_error", [
    ("Pendulum-Ignition-PyTest-v0", "Pendulum-PyBullet-PyTest-v0", 0.05),
    ("CartPoleDiscrete-Ignition-PyTest-v0", "CartPoleDiscrete-PyBullet-PyTest-v0", 0.03),
])
def test_compare_environments(env_name_a: str, env_name_b: str, max_error: float):
    template_compare_environments(env_name_a, env_name_b, max_error)



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest
pytestmark = pytest.mark.scenario

from ..common import utils
from scenario import gazebo as scenario
from ..common.utils import gazebo_fixture as gazebo

# Set the verbosity
scenario.set_verbosity(scenario.Verbosity_debug)


@pytest.mark.parametrize("gazebo",
                         [(0.001, 1.0, 1)],
                         indirect=True,
                         ids=utils.id_gazebo_fn)
def test_insert_multiple_worlds(gazebo: scenario.GazeboSimulator):

    empty_world_sdf = utils.get_empty_world_sdf()
    assert gazebo.insert_world_from_sdf(empty_world_sdf, "myWorld1")

    assert not gazebo.insert_world_from_sdf(empty_world_sdf, "myWorld1")
    assert gazebo.insert_world_from_sdf(empty_world_sdf, "myWorld2")

    assert gazebo.initialize()
    assert "myWorld1" in gazebo.world_names()
    assert "myWorld2" in gazebo.world_names()

    my_world1 = gazebo.get_world("myWorld1")
    my_world2 = gazebo.get_world("myWorld2")

    assert my_world1.id() != my_world2.id()
    assert my_world1.name() == "myWorld1"
    assert my_world2.name() == "myWorld2"


@pytest.mark.parametrize("gazebo",
                         [(0.001, 1.0, 1)],
                         indirect=True,
                         ids=utils.id_gazebo_fn)
def test_insert_multiple_world(gazebo: scenario.GazeboSimulator):

    multi_world_sdf = utils.get_multi_world_sdf_file()

    assert gazebo.insert_worlds_from_sdf(multi_world_sdf)
    assert gazebo.initialize()

    assert "world1" in gazebo.world_names()
    assert "world2" in gazebo.world_names()

    world1 = gazebo.get_world("world1")
    world2 = gazebo.get_world("world2")

    assert world1.name() == "world1"
    assert world2.name() == "world2"

    assert world1.id() != world2.id()


@pytest.mark.parametrize("gazebo",
                         [(0.001, 1.0, 1)],
                         indirect=True,
                         ids=utils.id_gazebo_fn)
def test_insert_multiple_world_rename(gazebo: scenario.GazeboSimulator):

    multi_world_sdf = utils.get_multi_world_sdf_file()

    assert not gazebo.insert_worlds_from_sdf(multi_world_sdf, ["only_one_name"])
    assert gazebo.insert_worlds_from_sdf(multi_world_sdf, ["myWorld1", "myWorld2"])
    assert gazebo.initialize()

    assert "myWorld1" in gazebo.world_names()
    assert "myWorld2" in gazebo.world_names()

    world1 = gazebo.get_world("myWorld1")
    world2 = gazebo.get_world("myWorld2")

    assert world1.name() == "myWorld1"
    assert world2.name() == "myWorld2"

    assert world1.id() != world2.id()


@pytest.mark.parametrize("gazebo",
                         [(0.001, 1.0, 1)],
                         indirect=True,
                         ids=utils.id_gazebo_fn)
def test_insert_world_multiple_calls(gazebo: scenario.GazeboSimulator):

    single_world_sdf = utils.get_empty_world_sdf()

    assert gazebo.insert_world_from_sdf(single_world_sdf)
    assert not gazebo.insert_world_from_sdf(single_world_sdf)
    assert gazebo.insert_world_from_sdf(single_world_sdf, "world2")
    assert gazebo.initialize()

    assert "default" in gazebo.world_names()
    assert "world2" in gazebo.world_names()

    world1 = gazebo.get_world("default")
    world2 = gazebo.get_world("world2")

    assert world1.name() == "default"
    assert world2.name() == "world2"

    assert world1.id() != world2.id()



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import numpy as np
from typing import List, Tuple
import idyntree.bindings as idt
from scenario import core as scenario_core
from gym_ignition.rbd import conversions
from gym_ignition.rbd.idyntree import numpy
from .helpers import FrameVelocityRepresentation, iDynTreeHelpers


class KinDynComputations:

    def __init__(self,
                 model_file: str,
                 considered_joints: List[str] = None,
                 world_gravity: np.ndarray = np.array([0, 0, -9.806]),
                 velocity_representation: FrameVelocityRepresentation =
                    FrameVelocityRepresentation.MIXED_REPRESENTATION) -> None:

        self.velocity_representation = velocity_representation

        self.kindyn = iDynTreeHelpers.get_kindyncomputations(model_file,
                                                             considered_joints,
                                                             velocity_representation)

        self.world_gravity = np.array(world_gravity)
        self.dofs = self.kindyn.getNrOfDegreesOfFreedom()

        if considered_joints is None:

            model: idt.Model = self.kindyn.model()

            all_joints = [model.getJointName(i) for i in range(model.getNrOfJoints())]
            self._considered_joints = all_joints
        else:
            self._considered_joints = considered_joints

    def joint_serialization(self) -> List[str]:

        return self._considered_joints

    def set_robot_state(self,
                        s: np.ndarray,
                        ds: np.ndarray,
                        world_H_base: np.ndarray = np.eye(4),
                        base_velocity: np.ndarray = np.zeros(6),
                        world_gravity: np.ndarray = None) -> None:

        gravity = world_gravity if world_gravity is not None else self.world_gravity

        if s.size != self.dofs:
            raise ValueError(s)

        if ds.size != self.dofs:
            raise ValueError(ds)

        if gravity.size != 3:
            raise ValueError(gravity)

        if world_H_base.shape != (4, 4):
            raise ValueError(world_H_base)

        if base_velocity.size != 6:
            raise ValueError(base_velocity)

        s_idyntree = numpy.FromNumPy.to_idyntree_dyn_vector(array=s)
        ds_idyntree = numpy.FromNumPy.to_idyntree_dyn_vector(array=ds)

        world_gravity_idyntree = \
            numpy.FromNumPy.to_idyntree_fixed_vector(array=gravity)

        world_H_base_idyntree = numpy.FromNumPy.to_idyntree_transform(
            position=world_H_base[0:3, 3], rotation=world_H_base[0:3, 0:3])

        base_velocity_idyntree = numpy.FromNumPy.to_idyntree_twist(
            linear_velocity=base_velocity[0:3], angular_velocity=base_velocity[3:6])

        ok_state = self.kindyn.setRobotState(world_H_base_idyntree,
                                             s_idyntree,
                                             base_velocity_idyntree,
                                             ds_idyntree,
                                             world_gravity_idyntree)

        if not ok_state:
            raise RuntimeError("Failed to set the robot state")

    def set_robot_state_from_model(self,
                                   model: scenario_core.Model,
                                   world_gravity: np.ndarray = None) \
            -> None:

        s = np.array(model.joint_positions(self.joint_serialization()))
        ds = np.array(model.joint_velocities(self.joint_serialization()))

        world_o_base = np.array(model.base_position())
        world_quat_base = np.array(model.base_orientation())

        # Velocity representations
        body = FrameVelocityRepresentation.BODY_FIXED_REPRESENTATION.to_idyntree()
        mixed = FrameVelocityRepresentation.MIXED_REPRESENTATION.to_idyntree()

        if self.kindyn.getFrameVelocityRepresentation() is mixed:

            base_linear_velocity = np.array(model.base_world_linear_velocity())
            base_angular_velocity = np.array(model.base_world_angular_velocity())

        elif self.kindyn.getFrameVelocityRepresentation() is body:

            base_linear_velocity = np.array(model.base_body_linear_velocity())
            base_angular_velocity = np.array(model.base_body_angular_velocity())

        else:
            raise RuntimeError("INERTIAL_FIXED_REPRESENTATION not yet supported")

        # Pack the data structures
        world_H_base = conversions.Transform.from_position_and_quaternion(
            position=world_o_base, quaternion=world_quat_base)
        base_velocity_6d = np.concatenate((base_linear_velocity, base_angular_velocity))

        self.set_robot_state(s=s,
                             ds=ds,
                             world_H_base=world_H_base,
                             base_velocity=base_velocity_6d,
                             world_gravity=world_gravity)

    def get_floating_base(self) -> str:

        return self.kindyn.getFloatingBase()

    def get_joint_positions(self) -> np.ndarray:

        vector = idt.VectorDynSize()

        if not self.kindyn.getJointPos(vector):
            raise RuntimeError("Failed to extract joint positions")

        return vector.toNumPy()

    def get_joint_velocities(self) -> np.ndarray:

        vector = idt.VectorDynSize()

        if not self.kindyn.getJointVel(vector):
            raise RuntimeError("Failed to extract joint velocities")

        return vector.toNumPy()

    def get_model_velocity(self) -> np.ndarray:

        nu = idt.VectorDynSize()

        if not self.kindyn.getModelVel(nu):
            raise RuntimeError("Failed to get the model velocity")

        return nu.toNumPy()

    def get_model_position(self) -> np.ndarray:

        W_H_B: idt.Transform = self.kindyn.getWorldBaseTransform()
        rpy: idt.Vector3 = self.kindyn.getWorldBaseTransform().getRotation().asRPY()

        q_base = np.concatenate([W_H_B.getPosition().toNumPy(), rpy.toNumPy()])

        return np.concatenate([q_base, self.get_joint_positions()])

    def get_world_transform(self, frame_name: str) -> np.ndarray:

        if self.kindyn.getFrameIndex(frame_name) < 0:
            raise ValueError(f"Frame '{frame_name}' does not exist")

        H = self.kindyn.getWorldTransform(frame_name)

        return numpy.ToNumPy.from_idyntree_transform(transform=H)

    def get_relative_transform(self,
                               ref_frame_name: str,
                               frame_name: str) -> np.ndarray:

        if self.kindyn.getFrameIndex(ref_frame_name) < 0:
            raise ValueError(f"Frame '{ref_frame_name}' does not exist")

        if self.kindyn.getFrameIndex(frame_name) < 0:
            raise ValueError(f"Frame '{frame_name}' does not exist")

        ref_H_other: idt.Transform = self.kindyn.getRelativeTransform(ref_frame_name,
                                                                      frame_name)

        return ref_H_other.asHomogeneousTransform().toNumPy()

    def get_world_base_transform(self) -> np.ndarray:

        W_H_B: idt.Transform = self.kindyn.getWorldBaseTransform()
        return W_H_B.asHomogeneousTransform().toNumPy()

    def get_relative_transform_explicit(self,
                                        ref_frame_origin: str,
                                        ref_frame_orientation: str,
                                        frame_origin: str,
                                        frame_orientation: str) -> np.ndarray:

        for frame in {ref_frame_origin,
                      ref_frame_orientation,
                      frame_origin,
                      frame_orientation}:

            if frame != "world" and self.kindyn.getFrameIndex(frame) < 0:
                raise ValueError(f"Frame '{frame}' does not exist")

        # Compute the transform AB_H_CD
        frameA = ref_frame_origin
        frameB = ref_frame_orientation
        frameC = frame_origin
        frameD = frame_orientation

        if frame_orientation == "world":
            frameD = frameC

        if ref_frame_orientation == "world":
            frameB = frameA

        frameA_index = self.kindyn.getFrameIndex(frameName=frameA)
        frameB_index = self.kindyn.getFrameIndex(frameName=frameB)
        frameC_index = self.kindyn.getFrameIndex(frameName=frameC)
        frameD_index = self.kindyn.getFrameIndex(frameName=frameD)

        ref_H_other: idt.Transform = self.kindyn.getRelativeTransformExplicit(
            frameA_index,
            frameB_index,
            frameC_index,
            frameD_index)

        AB_H_CD = ref_H_other

        if frame_orientation == "world":
            AB_H_C = AB_H_CD
            C_H_CW: idt.Transform = self.kindyn.getWorldTransform(frameC).inverse()
            C_H_CW.setPosition(idt.Position(0, 0, 0))
            AB_H_CW = AB_H_C * C_H_CW
            AB_H_CD = AB_H_CW

        if ref_frame_orientation == "world":
            A_H_CD = AB_H_CD
            AW_H_A: idt.Transform = self.kindyn.getWorldTransform(frameA)
            AW_H_A.setPosition(idt.Position(0, 0, 0))
            AW_H_CD = AW_H_A * A_H_CD
            AB_H_CD = AW_H_CD

        return AB_H_CD.asHomogeneousTransform().toNumPy()

    def get_relative_adjoint_wrench_transform_explicit(self,
                                                       ref_frame_origin: str,
                                                       ref_frame_orientation: str,
                                                       frame_origin: str,
                                                       frame_orientation: str) \
            -> np.ndarray:

        AB_H_CD = self.get_relative_transform_explicit(
            ref_frame_origin=ref_frame_origin,
            ref_frame_orientation=ref_frame_orientation,
            frame_origin=frame_origin,
            frame_orientation=frame_orientation)

        return numpy.FromNumPy.to_idyntree_transform(
            position=AB_H_CD[0:3, 3], rotation=AB_H_CD[0:3, 0:3]) \
            .asAdjointTransformWrench().toNumPy()

    def get_mass_matrix(self) -> np.ndarray:

        M = idt.MatrixDynSize()

        if not self.kindyn.getFreeFloatingMassMatrix(M):
            raise RuntimeError("Failed to get the free floating mass matrix")

        return M.toNumPy()

    def get_generalized_gravity_forces(self) -> np.ndarray:

        g = idt.FreeFloatingGeneralizedTorques(self.kindyn.model())

        if not self.kindyn.generalizedGravityForces(g):
            raise RuntimeError("Failed to get the generalized gravity forces")

        base_wrench: idt.Wrench = g.baseWrench()
        joint_torques: idt.JointDOFsDoubleArray = g.jointTorques()

        return np.concatenate([base_wrench.toNumPy().flatten(),
                               joint_torques.toNumPy().flatten()])

    def get_bias_forces(self) -> np.ndarray:

        h = idt.FreeFloatingGeneralizedTorques(self.kindyn.model())

        if not self.kindyn.generalizedBiasForces(h):
            raise RuntimeError("Failed to get the generalized bias forces")

        base_wrench: idt.Wrench = h.baseWrench()
        joint_torques: idt.JointDOFsDoubleArray = h.jointTorques()

        return np.concatenate([base_wrench.toNumPy().flatten(),
                               joint_torques.toNumPy().flatten()])

    def get_momentum(self) -> Tuple[np.ndarray, np.ndarray]:

        spatial_momentum = self.kindyn.getLinearAngularMomentum()
        momentum_6d = spatial_momentum.asVector().toNumPy()

        linear, angular = np.split(momentum_6d, 2)
        return linear, angular

    def get_centroidal_momentum(self) -> Tuple[np.ndarray, np.ndarray]:

        spatial_momentum = self.kindyn.getCentroidalTotalMomentum()
        momentum_6d = spatial_momentum.asVector().toNumPy()

        linear, angular = np.split(momentum_6d, 2)
        return linear, angular

    def get_com_position(self) -> np.ndarray:

        W_p_com = self.kindyn.getCenterOfMassPosition()
        return W_p_com.toNumPy()

    def get_com_velocity(self) -> np.ndarray:

        # Velocity representations
        body = FrameVelocityRepresentation.BODY_FIXED_REPRESENTATION.to_idyntree()
        mixed = FrameVelocityRepresentation.MIXED_REPRESENTATION.to_idyntree()

        if self.kindyn.getFrameVelocityRepresentation() is mixed:

            # The method always returns the MIXED velocity of the CoM, regardless of
            # how KinDynComputations was configured.
            v_com = self.kindyn.getCenterOfMassVelocity()
            return v_com.toNumPy()

        elif self.kindyn.getFrameVelocityRepresentation() is body:

            # Get the transform of the base frame
            W_H_B = self.kindyn.getWorldBaseTransform()
            _, W_R_B = numpy.ToNumPy.from_idyntree_transform(transform=W_H_B, split=True)

            # Get the rotation between world and base frame
            B_R_W = np.linalg.inv(W_R_B)

            # Convert linear velocity from MIXED to BODY representation
            BW_v_com = self.kindyn.getCenterOfMassVelocity().toNumPy()
            B_v_com = B_R_W @ BW_v_com

            return B_v_com

        else:
            raise RuntimeError("INERTIAL_FIXED_REPRESENTATION not yet supported")

    def get_average_velocity(self) -> np.ndarray:

        twist: idt.Twist = self.kindyn.getAverageVelocity()
        return twist.toNumPy()

    def get_centroidal_average_velocity(self) -> np.ndarray:

        twist: idt.Twist = self.kindyn.getCentroidalAverageVelocity()
        return twist.toNumPy()

    def get_frame_jacobian(self, frame_name: str) -> np.ndarray:

        if self.kindyn.getFrameIndex(frame_name) < 0:
            raise ValueError(f"Frame '{frame_name}' does not exist")

        J = idt.MatrixDynSize(6, self.dofs + 6)

        if not self.kindyn.getFrameFreeFloatingJacobian(frame_name, J):
            raise RuntimeError("Failed to get the frame free floating jacobian")

        return J.toNumPy()

    def get_linear_angular_momentum_jacobian(self) -> np.ndarray:

        J_mom = idt.MatrixDynSize()

        if not self.kindyn.getLinearAngularMomentumJacobian(J_mom):
            raise RuntimeError("Failed to get the momentum jacobian")

        return J_mom.toNumPy()

    def get_centroidal_total_momentum_jacobian(self) -> np.ndarray:

        J_cmm = idt.MatrixDynSize()

        if not self.kindyn.getCentroidalTotalMomentumJacobian(J_cmm):
            raise RuntimeError("Failed to get the centroidal total momentum jacobian")

        return J_cmm.toNumPy()

    def get_average_velocity_jacobian(self) -> np.ndarray:

        J_avg_vel = idt.MatrixDynSize()

        if not self.kindyn.getAverageVelocityJacobian(J_avg_vel):
            raise RuntimeError("Failed to get the average velocity jacobian")

        return J_avg_vel.toNumPy()

    def get_centroidal_average_velocity_jacobian(self) -> np.ndarray:

        J_cen_avg_vel = idt.MatrixDynSize()

        if not self.kindyn.getCentroidalAverageVelocityJacobian(J_cen_avg_vel):
            raise RuntimeError("Failed to get the average velocity jacobian")

        return J_cen_avg_vel.toNumPy()

    def get_frame_bias_acc(self, frame_name: str) -> np.ndarray:

        if self.kindyn.getFrameIndex(frame_name) < 0:
            raise ValueError(f"Frame '{frame_name}' does not exist")

        dJ_nu = self.kindyn.getFrameBiasAcc(frame_name)

        return dJ_nu.toNumPy()

    def get_com_bias_acc(self) -> np.ndarray:

        dJ_nu = self.kindyn.getCenterOfMassBiasAcc()
        return dJ_nu.toNumPy()



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
import gym_ignition.base.task
from scenario import core as scenario_core


class TaskRandomizer(abc.ABC):

    @abc.abstractmethod
    def randomize_task(self,
                       task: gym_ignition.base.task.Task,
                       **kwargs) -> None:
        """
        Randomize a :py:class:`~gym_ignition.base.task.Task` instance.

        Args:
            task: the task to randomize.

        Note:
            Note that each task has a :py:attr:`~gym_ignition.base.task.Task.world`
            property that provides access to the simulated
            :py:class:`scenario.bindings.core.World`.
        """
        pass


class PhysicsRandomizer(abc.ABC):
    """
    Abstract class that provides the machinery for randomizing the physics of a Task.

    Args:
        randomize_after_rollouts_num: defines after many rollouts physics should be
            randomized (i.e. the amount of times :py:meth:`gym.Env.reset` is called).
    """

    def __init__(self, randomize_after_rollouts_num: int = 0):

        self._rollout_counter = randomize_after_rollouts_num
        self.randomize_after_rollouts_num = randomize_after_rollouts_num

    @abc.abstractmethod
    def randomize_physics(self, task: gym_ignition.base.task.Task, **kwargs) -> None:
        """
        Method that insert and configures the physics of a Task's world.

        By default this method loads a plugin that uses DART with no randomizations.
        Randomizing physics engine parameters or changing physics engine backend could be
        done by redefining this method and passing it to
        :py:class:`~gym_ignition.runtimes.gazebo_runtime.GazeboRuntime`.

        Args:
            task: A task containing a world object without physics.
        """
        pass

    @abc.abstractmethod
    def get_engine(self):
        """
        Return the physics engine to use for the rollout.

        Note:

            Supported physics engines:

            - :py:const:`scenario.bindings.gazebo.PhysicsEngine_dart`

        Return:
            The desired physics engine to set in the world.
        """

        pass

    def increase_rollout_counter(self) -> None:
        """
        Increase the rollouts counter.
        """

        if self.randomize_after_rollouts_num != 0:
            assert self._rollout_counter != 0
            self._rollout_counter -= 1

    def physics_expired(self) -> bool:
        """
        Checks if the physics needs to be randomized.

        Return:
            True if the physics has expired, false otherwise.
        """

        if self.randomize_after_rollouts_num == 0:
            return False

        if self._rollout_counter == 0:
            self._rollout_counter = self.randomize_after_rollouts_num
            return True

        return False


class ModelRandomizer(abc.ABC):

    @abc.abstractmethod
    def randomize_model(self, task: gym_ignition.base.task.Task, **kwargs) \
            -> scenario_core.Model:
        """
        Randomize the model.

        Args:
            task: The task that operates on the model to randomize.

        Return:
            The randomized model.
        """
        pass


class ModelDescriptionRandomizer(abc.ABC):

    @abc.abstractmethod
    def randomize_model_description(self, task: gym_ignition.base.task.Task, **kwargs) \
            -> str:
        """
        Randomize the model description.

        Args:
            task: The task that operates on the model description to randomize.

        Return:
            A string with the randomized model description.
        """
        pass



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import pytest

pytestmark = pytest.mark.scenario

from ..common import utils
from scenario import core
from scenario import gazebo as scenario
from ..common.utils import gazebo_fixture as gazebo

from pathlib import Path

# Set the verbosity
scenario.set_verbosity(scenario.Verbosity_debug)


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_download_model_from_fuel(gazebo: scenario.GazeboSimulator):

    assert gazebo.initialize()

    # Get the default world
    world = gazebo.get_world()

    # Download a model from Fuel (testing a name with spaces)
    model_name = "Electrical Box"
    model_sdf = scenario.get_model_file_from_fuel(
        f"https://fuel.ignitionrobotics.org/openrobotics/models/{model_name}", False
    )
    assert model_sdf

    assert world.insert_model(model_sdf, core.Pose_identity())
    assert model_name in world.model_names()

    # Insert another model changing its name
    other_model_name = "my_box"
    other_model_pose = core.Pose([3.0, 0.0, 0.0], [1.0, 0, 0, 0])
    assert world.insert_model(model_sdf, other_model_pose, other_model_name)
    assert other_model_name in world.model_names()

    assert gazebo.run()


@pytest.mark.parametrize(
    "gazebo", [(0.001, 1.0, 1)], indirect=True, ids=utils.id_gazebo_fn
)
def test_fuel_world(gazebo):
    # (setup) load a world that includes a fuel model
    worlds_folder = Path(__file__) / ".." / ".." / "assets" / "worlds"
    world_file = worlds_folder / "fuel_support.sdf"
    assert gazebo.insert_world_from_sdf(str(world_file.resolve()))
    assert gazebo.initialize()
    assert gazebo.run(paused=True)

    # the actual test
    assert "ground_plane" in gazebo.get_world().model_names()



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

# Pendulum
from . import pendulum_swingup

# Cartpole discrete
from . import cartpole_discrete_balancing

# Cartpole continuous
from . import cartpole_continuous_swingup
from . import cartpole_continuous_balancing



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import gym_ignition_models
from gym_ignition import base, utils
from gym_ignition.base import runtime
from gym_ignition.utils import logger
from gym_ignition.utils.typing import *
from scenario import gazebo as scenario


class GazeboRuntime(runtime.Runtime):
    """
    Implementation of :py:class:`~gym_ignition.base.runtime.Runtime` for the Ignition
    Gazebo simulator.

    Args:
        task_cls: The class of the handled task.
        agent_rate: The rate at which the environment is called.
        physics_rate: The rate of the physics engine.
        real_time_factor: The desired RTF of the simulation.
        physics_engine: *(optional)* The physics engine to use.
        world: *(optional)* The path to an SDF world file. The world should not contain
            any physics plugin.

    Note:
        Physics randomization is still experimental and it could change in the future.
        Physics is loaded only once, when the simulator starts. In order to change the
        physics, a new simulator should be created.
    """

    metadata = {'render.modes': ['human']}

    def __init__(self,
                 task_cls: type,
                 agent_rate: float,
                 physics_rate: float,
                 real_time_factor: float,
                 physics_engine = scenario.PhysicsEngine_dart,
                 world: str = None,
                 **kwargs):

        # Gazebo attributes
        self._gazebo = None
        self._physics_rate = physics_rate
        self._real_time_factor = real_time_factor

        # Store the desired physics engine
        self._physics_engine = physics_engine

        # World attributes
        self._world = None
        self._world_sdf = world
        self._world_name = None

        # Create the Task object
        task = task_cls(agent_rate=agent_rate, **kwargs)

        if not isinstance(task, base.task.Task):
            raise RuntimeError("The task is not compatible with the runtime")

        # Wrap the task with the runtime
        super().__init__(task=task, agent_rate=agent_rate)

        # Trigger the initialization of the simulator and the world
        _ = self.gazebo

        # Initialize the spaces
        self.action_space, self.observation_space = self.task.create_spaces()

        # Store the spaces also in the task
        self.task.action_space = self.action_space
        self.task.observation_space = self.observation_space

        # Seed the environment
        self.seed()

    # =================
    # Runtime interface
    # =================

    def timestamp(self) -> float:

        return self.world.time()

    # =================
    # gym.Env interface
    # =================

    def step(self, action: Action) -> State:

        if not self.action_space.contains(action):
            logger.warn("The action does not belong to the action space")

        # Set the action
        self.task.set_action(action)

        # Step the simulator
        ok_gazebo = self.gazebo.run()
        assert ok_gazebo, "Failed to step gazebo"

        # Get the observation
        observation = self.task.get_observation()
        assert isinstance(observation, np.ndarray)

        if not self.observation_space.contains(observation):
            logger.warn("The observation does not belong to the observation space")

        # Get the reward
        reward = self.task.get_reward()
        assert isinstance(reward, float), "Failed to get the reward"

        # Check termination
        done = self.task.is_done()

        # Get info
        info = self.task.get_info()

        return State((Observation(observation), Reward(reward), Done(done), Info(info)))

    def reset(self) -> Observation:

        # Reset the task
        self.task.reset_task()

        # Execute a paused step
        ok_run = self.gazebo.run(paused=True)

        if not ok_run:
            raise RuntimeError("Failed to run Gazebo")

        # Get the observation
        observation = self.task.get_observation()
        assert isinstance(observation, np.ndarray)

        if not self.observation_space.contains(observation):
            logger.warn("The observation does not belong to the observation space")

        return Observation(observation)

    def render(self, mode: str = 'human', **kwargs) -> None:

        if mode != 'human':
            raise ValueError(f"Render mode '{mode}' not supported")

        gui_ok = self.gazebo.gui()

        if not gui_ok:
            raise RuntimeError("Failed to render the environment")

        return

    def close(self) -> None:

        ok_close = self.gazebo.close()

        if not ok_close:
            raise RuntimeError("Failed to close Gazebo")

    def seed(self, seed: int = None) -> SeedList:

        # This method also seeds the spaces. To create them, the task could use the world.
        # Here we check that is has been created.
        if not self.task.has_world():
            raise RuntimeError("The world has never been created")

        # Seed the task (it will also seed the spaces)
        seed = self.task.seed_task(seed)

        return seed

    # ==============================
    # Properties and Private Methods
    # ==============================

    @property
    def gazebo(self) -> scenario.GazeboSimulator:

        if self._gazebo is not None:
            assert self._gazebo.initialized()
            return self._gazebo

        # Compute the number of physics iteration to execute at every environment step
        num_of_steps_per_run = self._physics_rate / self.agent_rate

        if num_of_steps_per_run != int(num_of_steps_per_run):
            logger.warn("Rounding the number of iterations to {} from the nominal {}"
                        .format(int(num_of_steps_per_run), num_of_steps_per_run))

        # Create the simulator
        gazebo = scenario.GazeboSimulator(1.0 / self._physics_rate,
                                          self._real_time_factor,
                                          int(num_of_steps_per_run))

        # Store the simulator
        self._gazebo = gazebo

        # Insert the world (it will initialize the simulator)
        _ = self.world
        assert self._gazebo.initialized()

        return self._gazebo

    @property
    def world(self) -> scenario.World:

        if self._world is not None:
            assert self.gazebo.initialized()
            return self._world

        if self._gazebo is None:
            raise RuntimeError("Gazebo has not yet been created")

        # Help type hinting
        self._gazebo: scenario.GazeboSimulator

        if self._gazebo.initialized():
            raise RuntimeError("Gazebo was already initialized, cannot insert world")

        if self._world_sdf is None:
            self._world_sdf = ""
            self._world_name = utils.scenario.get_unique_world_name("default")
        else:
            sdf_world_name = scenario.get_world_name_from_sdf(self._world_sdf)
            self._world_name = utils.scenario.get_unique_world_name(sdf_world_name)

        # Load the world
        ok_world = self._gazebo.insert_world_from_sdf(self._world_sdf, self._world_name)

        if not ok_world:
            raise RuntimeError("Failed to load SDF world")

        # Initialize the simulator
        ok_initialize = self._gazebo.initialize()

        if not ok_initialize:
            raise RuntimeError("Failed to initialize Gazebo")

        if not self._gazebo.initialized():
            raise RuntimeError("Gazebo was not initialized")

        # Get the world
        world = self._gazebo.get_world(self._world_name)

        assert self._world_sdf is not None
        assert self._world_name in self._gazebo.world_names()

        if self._world_sdf == "":

            # Insert the ground plane
            ok_ground = world.insert_model(
                gym_ignition_models.get_model_file("ground_plane"))

            if not ok_ground:
                raise RuntimeError("Failed to insert the ground plane")

        # Set the world in the task
        self.task.world = world

        # Select the physics engine
        world.set_physics_engine(engine=self._physics_engine)

        # Store the world
        self._world = world

        return self._world



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
from typing import List
from gym_ignition import scenario
from scenario import core as scenario_core
from gym_ignition.utils.scenario import get_unique_model_name


class ICubGazeboABC(scenario.model_wrapper.ModelWrapper,
                    abc.ABC):

    DOFS = 32
    NUM_LINKS = 39
    NUM_JOINTS = 32

    initial_positions = {
        # Left leg
        'l_knee': -1.05,
        'l_ankle_pitch': -0.57, 'l_ankle_roll': -0.024,
        'l_hip_pitch': 0.48, 'l_hip_roll': 0.023, 'l_hip_yaw': -0.005,
        # Left arm
        'l_elbow': 0.54,
        'l_wrist_pitch': 0.0, 'l_wrist_prosup': 0.0, 'l_wrist_yaw': 0.0,
        'l_shoulder_pitch': -0.159, 'l_shoulder_roll': 0.435, 'l_shoulder_yaw': 0.183,
        # Head
        'neck_pitch': 0.0, 'neck_roll': 0.0, 'neck_yaw': 0.0,
        # Right leg
        'r_knee': -1.05,
        'r_ankle_pitch': -0.57, 'r_ankle_roll': -0.024,
        'r_hip_pitch': 0.48, 'r_hip_roll': 0.023, 'r_hip_yaw': -0.005,
        # Right arm
        'r_elbow': 0.54,
        'r_wrist_pitch': 0.0, 'r_wrist_prosup': 0.0, 'r_wrist_yaw': 0.0,
        'r_shoulder_pitch': -0.159, 'r_shoulder_roll': 0.435, 'r_shoulder_yaw': 0.183,
        # Torso
        'torso_pitch': 0.1, 'torso_roll': 0.0, 'torso_yaw': 0.0,
    }

    def __init__(self,
                 world: scenario_core.World,
                 position: List[float],
                 orientation: List[float],
                 model_file: str = None):

        # Get a unique model name
        model_name = get_unique_model_name(world, "icub")

        # Initial pose
        initial_pose = scenario_core.Pose(position, orientation)

        # Get the model file (URDF or SDF) and allow to override it
        if model_file is None:
            model_file = self.get_model_file()

        # Insert the model
        ok_model = world.to_gazebo().insert_model(model_file,
                                                  initial_pose,
                                                  model_name)

        if not ok_model:
            raise RuntimeError("Failed to insert model")

        # Get the model
        model = world.get_model(model_name)

        # Initialize base class
        super().__init__(model=model)

        # Store the initial positions
        q0 = list(self.initial_positions.values())
        joint_names = list(self.initial_positions.keys())
        assert self.dofs() == len(q0) == len(joint_names)

        ok_q0 = self.to_gazebo().reset_joint_positions(q0, joint_names)
        assert ok_q0, "Failed to set initial position"


class ICubGazebo(ICubGazeboABC,
                 scenario.model_with_file.ModelWithFile):

    def __init__(self,
                 world: scenario_core.World,
                 position: List[float] = (0., 0., 0.572),
                 orientation: List[float] = (0, 0, 0, 1.0),
                 model_file: str = None):

        super().__init__(world=world,
                         position=position,
                         orientation=orientation,
                         model_file=model_file)

    @classmethod
    def get_model_file(cls) -> str:

        import gym_ignition_models
        return gym_ignition_models.get_model_file("iCubGazeboV2_5")


class ICubGazeboSimpleCollisions(ICubGazeboABC):

    def __init__(self,
                 world: scenario_core.World,
                 position: List[float] = (0., 0., 0.572),
                 orientation: List[float] = (0, 0, 0, 1.0),
                 model_file: str = None):

        super().__init__(world=world,
                         position=position,
                         orientation=orientation,
                         model_file=model_file)

    @classmethod
    def get_model_file(cls) -> str:

        import gym_ignition_models
        return gym_ignition_models.get_model_file("iCubGazeboSimpleCollisionsV2_5")



import abc
import numpy as np
from gym_ignition import rbd
from typing import Tuple, Union
import idyntree.bindings as idt


class FromNumPy(abc.ABC):

    @staticmethod
    def to_idyntree_dyn_vector(array: np.ndarray) -> idt.VectorDynSize:

        return idt.VectorDynSize_FromPython(array)

    @staticmethod
    def to_idyntree_fixed_vector(array: np.ndarray):

        size = array.size
        supported_sizes = [3, 4, 6]

        if size not in supported_sizes:
            raise ValueError(array)

        if size == 3:
            idyntree_vector = idt.Vector3()
        elif size == 4:
            idyntree_vector = idt.Vector4()
        elif size == 6:
            idyntree_vector = idt.Vector6()
        else:
            raise RuntimeError

        return idyntree_vector.FromPython(array)

    @staticmethod
    def to_idyntree_position(position: np.ndarray) -> idt.Position:

        if position.size != 3:
            raise ValueError("The position array must have 3 elements")

        return idt.Position(position[0], position[1], position[2])

    @staticmethod
    def to_idyntree_rotation(quaternion: np.ndarray) -> idt.Rotation:

        if quaternion.size != 4:
            raise ValueError("The quaternion array must have 4 elements")

        quat = idt.Vector4_FromPython(quaternion)

        R = idt.Rotation()
        R.fromQuaternion(quat)

        return R

    @staticmethod
    def to_idyntree_transform(position: np.ndarray,
                              quaternion: np.ndarray = None,
                              rotation: np.ndarray = None) -> idt.Transform:

        if quaternion is None and rotation is None:
            raise ValueError("You must pass either a quaternion or a rotation")

        if quaternion is not None and rotation is not None:
            raise ValueError("You must pass either a quaternion or a rotation")

        if rotation is not None:
            quaternion = rbd.conversions.Quaternion.from_matrix(matrix=rotation)

        p = FromNumPy.to_idyntree_position(position=position)
        R = FromNumPy.to_idyntree_rotation(quaternion=quaternion)

        H = idt.Transform()
        H.setPosition(p)
        H.setRotation(R)

        return H

    @staticmethod
    def to_idyntree_twist(linear_velocity: np.ndarray,
                          angular_velocity: np.ndarray) -> idt.Twist:

        if linear_velocity.size != 3:
            raise ValueError("The linear velocity must have 3 elements")

        if angular_velocity.size != 3:
            raise ValueError("The angular velocity must have 3 elements")

        twist_numpy = np.concatenate((linear_velocity, angular_velocity))

        twist = idt.Twist_FromPython(twist_numpy)

        return twist


class ToNumPy(abc.ABC):

    @staticmethod
    def from_idyntree_vector(vector) -> np.ndarray:

        input_types = (
            idt.Vector3,
            idt.Vector4,
            idt.Vector6,
            idt.VectorDynSize,
            idt.Matrix3x3,
            idt.Matrix4x4,
            idt.Matrix6x6,
            idt.MatrixDynSize,
        )

        if not isinstance(vector, input_types):
            raise ValueError(vector)

        return np.array(vector.toNumPy())

    @staticmethod
    def from_idyntree_transform(transform: idt.Transform, split: bool = False) \
            -> Union[Tuple[np.ndarray, np.ndarray], np.ndarray]:

        if not isinstance(transform, idt.Transform):
            raise ValueError(transform)

        rotation = transform.getRotation()
        position = transform.getPosition()

        if split:
            return position.toNumPy(), rotation.toNumPy()

        else:
            H = np.zeros(shape=[4, 4])
            H[0:3, 3] = position.toNumPy()
            H[0:3, 0:3] = rotation.toNumPy()

            return H



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import tempfile
import numpy as np
from . import utils


def test_contacts():
    # Get the simulator
    gazebo = utils.Gazebo(physics_rate=1000, iterations=1, rtf=100)

    # Create the first cube and insert it in the simulation
    cube1 = utils.CubeGazeboRobot(gazebo=gazebo.simulator,
                                  initial_position=np.array([0, 0, 1.0]))

    # Create the second cube and insert it in the simulation
    cube2 = utils.CubeGazeboRobot(gazebo=gazebo.simulator,
                                  initial_position=np.array([0, 0, 2.5]))

    # Execute the first simulation step
    gazebo.step()

    # The cubes should be falling without contacts
    assert len(cube1.links_in_contact()) == 0
    assert len(cube2.links_in_contact()) == 0

    # Perform 500 steps.
    for _ in range(500):
        gazebo.step()

    # Cube1 should be touching ground
    assert len(cube1.links_in_contact()) == 1
    assert cube1.links_in_contact()[0] == 'cube'
    contact_data1 = cube1.contact_data('cube')
    assert len(contact_data1) > 0
    assert contact_data1[0].bodyA == cube1.name() + "::cube_collision"
    assert contact_data1[0].bodyB == "ground_plane::collision"

    # Cube 2 should be still floating
    assert len(cube2.links_in_contact()) == 0
    assert len(cube2.contact_data('cube')) == 0

    # Perform 500 steps.
    for _ in range(500):
        gazebo.step()

    # Now cube2 should be in contact with cube1
    assert len(cube2.links_in_contact()) == 1
    assert cube2.links_in_contact()[0] == 'cube'
    contact_data2 = cube2.contact_data('cube')
    assert len(contact_data2) > 0
    assert contact_data2[0].bodyA == cube2.name() + "::cube_collision"
    assert contact_data2[0].bodyB == cube1.name() + "::cube_collision"

    # And cube1 should be in contact with cube2 and ground_plane
    assert len(cube1.links_in_contact()) == 1
    assert cube1.links_in_contact()[0] == 'cube'
    contact_data1 = cube1.contact_data('cube')
    assert len(contact_data1) == 2

    for contact in contact_data1:
        assert contact.bodyA == cube1.name() + "::cube_collision"
        assert contact.bodyB == cube2.name() + "::cube_collision" or \
               contact.bodyB == "ground_plane::collision"



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import os
from setuptools import setup, find_packages
from cmake_build_extension import BuildExtension, CMakeExtension

# Read the contents of your README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='gym-ignition',
    author="Diego Ferigo",
    author_email="diego.ferigo@iit.it",
    description="Gym-Ignition: A toolkit for developing OpenAI Gym environments "
                "simulated with Ignition Gazebo.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/robotology/gym-ignition',
    keywords="openai gym reinforcement learning environment gazebo robotics ignition",
    license="LGPL",
    platforms='any',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Operating System :: POSIX :: Linux",
        "Topic :: Games/Entertainment :: Simulation",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Framework :: Robot Framework",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Programming Language :: C++",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
    ],
    use_scm_version={
        'local_scheme': 'dirty-tag',
    },
    setup_requires=['setuptools_scm'],
    python_requires='>=3.8',
    install_requires=[
        'gym >= 0.13.1',
        'numpy',
        'gym_ignition_models',
        'lxml',
        'idyntree',
    ],
    extras_require=dict(
        website=[
            "sphinx",
            "sphinx-book-theme < 0.0.39",
            "sphinx-autodoc-typehints",
            "sphinx_fontawesome",
            "sphinx-multiversion",
            'sphinx-tabs',
            "breathe",
        ],
        test=[
            "pytest",
            "pytest-xvfb",
            "pytest-icdiff",
        ]
    ),
    packages=find_packages("python"),
    package_dir={'': "python"},
    ext_modules=[CMakeExtension(name='InstallAllTargets',
                                install_prefix="scenario",
                                cmake_build_type='PyPI',
                                cmake_depends_on=["idyntree"],
                                disable_editable=True)],
    cmdclass={
        'build_ext': BuildExtension,
    },
    zip_safe=False,
)



# Copyright (C) 2019 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import abc
import gym
import numpy as np
from scenario import core
from gym.utils import seeding
from typing import Dict, Tuple
from gym_ignition.utils.typing import ActionSpace, ObservationSpace
from gym_ignition.utils.typing import Action, Observation, Reward, SeedList


class Task(abc.ABC):
    """
    Interface to define a decision-making task.

    The Task is the central interface of each environment implementation.
    It defines the logic of the environment in a format that is agnostic of both the
    runtime (either simulated or real-time) and the models it operates on.

    :py:class:`~gym_ignition.base.runtime.Runtime` instances are the real objects returned
    to the users when they call :py:class:`gym.make`. Depending on the type of the
    runtime, it could contain one or more :py:class:`Task` objects.
    The :py:class:`~gym_ignition.base.runtime.Runtime` is a relay class that calls the
    logic of the :py:class:`Task` from its interface methods and implements the real
    :py:meth:`gym.Env.step`.
    Simulated runtimes step the physics engine, instead, real-time
    runtimes, enforce real-time execution.

    A :py:class:`Task` object is meant to be:

    - Independent from the selected :py:class:`~gym_ignition.base.runtime.Runtime`.
      In fact, it defines only the decision making logic;
    - Independent from the :py:class:`~scenario.core.Model` objects it operates on.
      This is achieved thanks to the model abstraction provided by
      :cpp:class:`scenario::core::Model`.

    The population of the world where the task operates is demanded to a
    :py:class:`gym.Wrapper` object, that acts as an environment randomizer.
    """

    action_space: gym.spaces.Space = None
    observation_space: gym.spaces.Space = None

    def __init__(self, agent_rate: float) -> None:

        # World object
        self._world = None

        #: Rate of the agent.
        #: It matches the rate at which the :py:class:`Gym.Env` methods are called.
        self.agent_rate = agent_rate

        #: RNG available to the object to ensure reproducibility.
        #: Use it for all the random resources.
        self.np_random: np.random.RandomState

        #: The seed of the task
        self.seed: int

        # Initialize the RNG and the seed
        self.np_random, self.seed = seeding.np_random()

    # ==========
    # PROPERTIES
    # ==========

    @property
    def world(self) -> core.World:
        """
        Get the world where the task is operating.

        Returns:
            The world object.
        """

        if self._world is not None:
            return self._world

        raise Exception("The world was never stored")

    @world.setter
    def world(self, world: core.World) -> None:

        if world is None or world.name == "":
            raise ValueError("World not valid")

        # Store the world
        self._world = world

    def has_world(self) -> bool:
        """
        Check if the world was stored.

        Returns:
            True if the task has a valid world, False otherwise.
        """

        return self._world is not None and self._world.name != ""

    # ==============
    # Task Interface
    # ==============

    @abc.abstractmethod
    def create_spaces(self) -> Tuple[ActionSpace, ObservationSpace]:
        """
        Create the action and observations spaces.

        Note:
            This method does not currently have access to the Models part of the
            environment. If the Task is meant to work on different models, we recommend
            using their URDF / SDF model to extract the information you need
            (e.g. number of DoFs, joint position limits, etc). Since actions and
            observations are often normalized, in many cases there's no need to extract
            a lot of information from the model file.

        Raises:
            RuntimeError: In case of failure.

        Returns:
            A tuple containing the action and observation spaces.
        """

    @abc.abstractmethod
    def reset_task(self) -> None:
        """
        Reset the task.

        This method contains the logic for resetting the task.
        It is called in the :py:meth:`gym.Env.reset` method of the corresponding
        environment.

        Raises:
            RuntimeError: In case of failure.
        """

    @abc.abstractmethod
    def set_action(self, action: Action) -> None:
        """
        Set the task action.

        This method contains the logic for setting the environment action.
        It is called in the beginning of the :py:meth:`gym.Env.step` method.

        Args:
            action: The action to set.

        Raises:
            RuntimeError: In case of failure.
        """

    @abc.abstractmethod
    def get_observation(self) -> Observation:
        """
        Return the task observation.

        This method contains the logic for constructing the environment observation.
        It is called in the end of both :py:meth:`gym.Env.reset` and
        :py:meth:`gym.Env.step` methods.

        Raises:
            RuntimeError: In case of failure.

        Returns:
            The task observation.
        """

    @abc.abstractmethod
    def get_reward(self) -> Reward:
        """
        Return the task reward.

        This method contains the logic for computing the environment reward.
        It is called in the end of the :py:meth:`gym.Env.step` method.

        Raises:
            RuntimeError: In case of failure.

        Returns:
            The scalar reward.
        """

    @abc.abstractmethod
    def is_done(self) -> bool:
        """
        Return the task termination flag.

        This method contains the logic for defining when the environment has terminated.
        Subsequent calls to :py:meth:`Task.set_action` should be preceded by a task
        reset through :py:meth:`Task.reset_task`.

        It is called in the end of the :py:meth:`gym.Env.step` method.

        Raises:
            RuntimeError: In case of failure.

        Returns:
            True if the environment terminated, False otherwise.
        """

    def get_info(self) -> Dict:
        """
        Return the info dictionary.

        Returns:
            A ``dict`` with extra information of the task.
        """
        return {}

    def seed_task(self, seed: int = None) -> SeedList:
        """
        Seed the task.

        This method configures the :py:attr:`Task.np_random` RNG.

        Args:
            seed: The seed number.

        Return:
            The list of seeds used by the task.
        """

        # Create the seed if not passed
        seed = np.random.randint(2**32 - 1) if seed is None else seed

        # Get an instance of the random number generator from gym utils.
        # This is necessary to have an independent rng for each environment.
        self.np_random, self.seed = seeding.np_random(seed)

        # Seed the spaces
        self.action_space.seed(self.seed)
        self.observation_space.seed(self.seed)

        return SeedList([self.seed])



# Copyright (C) 2020 Istituto Italiano di Tecnologia (IIT). All rights reserved.
# This software may be modified and distributed under the terms of the
# GNU Lesser General Public License v2.1 or any later version.

import os
import abc
from typing import List
from enum import Enum, auto
import idyntree.bindings as idt
from gym_ignition.utils import resource_finder


class FrameVelocityRepresentation(Enum):

    MIXED_REPRESENTATION = auto()
    BODY_FIXED_REPRESENTATION = auto()
    INERTIAL_FIXED_REPRESENTATION = auto()

    def to_idyntree(self):

        if self.value == FrameVelocityRepresentation.MIXED_REPRESENTATION.value:
            return idt.MIXED_REPRESENTATION
        elif self.value == FrameVelocityRepresentation.BODY_FIXED_REPRESENTATION.value:
            return idt.BODY_FIXED_REPRESENTATION
        elif self.value == FrameVelocityRepresentation.INERTIAL_FIXED_REPRESENTATION.value:
            return idt.INERTIAL_FIXED_REPRESENTATION
        else:
            raise ValueError(self.value)


class iDynTreeHelpers(abc.ABC):

    @staticmethod
    def get_model_loader(model_file: str, considered_joints: List[str] = None):

        # Find the urdf file
        urdf_file = resource_finder.find_resource(file_name=model_file)

        # Get the file extension
        folder, model_file = os.path.split(urdf_file)
        model_name, extension = os.path.splitext(model_file)

        if extension == ".sdf":
            raise RuntimeError("SDF models are not currently supported by iDynTree")

        # Create the model loader
        mdl_loader = idt.ModelLoader()

        # Load the urdf model
        if considered_joints is None:
            ok_load = mdl_loader.loadModelFromFile(urdf_file)
        else:
            ok_load = mdl_loader.loadReducedModelFromFile(urdf_file, considered_joints)

        if not ok_load:
            raise RuntimeError(f"Failed to load model from file '{urdf_file}'")

        return mdl_loader

    @staticmethod
    def get_kindyncomputations(
            model_file: str,
            considered_joints: List[str] = None,
            velocity_representation: FrameVelocityRepresentation =
                FrameVelocityRepresentation.MIXED_REPRESENTATION):

        # Get the model loader
        model_loader = iDynTreeHelpers.get_model_loader(model_file, considered_joints)

        # Create KinDynComputations and insert the model
        kindyn = idt.KinDynComputations()
        ok_load = kindyn.loadRobotModel(model_loader.model())

        if not ok_load:
            raise RuntimeError("Failed to load model")

        # Configure the velocity representation
        velocity_representation_idyntree = velocity_representation.to_idyntree()
        ok_repr = kindyn.setFrameVelocityRepresentation(velocity_representation_idyntree)

        if not ok_repr:
            raise RuntimeError("Failed to set the velocity representation")

        return kindyn
